@@ -20,26 +20,57 @@
 # IN THE SOFTWARE.
 
 import collections
+import deprecated
 import os
 import subprocess as sp
 import typing as t
 
 __author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
-__version__ = '0.2.0'
+__version__ = '0.4.1'
 
-Branch = collections.namedtuple('Branch', 'name,current')
-FileStatus = collections.namedtuple('FileStatus', 'mode,filename')
-RefWithSha = collections.namedtuple('RefWithSha', 'ref,sha')
+
+class Branch(t.NamedTuple):
+  name: str
+  current: bool
+
+
+class FileStatus(t.NamedTuple):
+  mode: str
+  filename: str
+
+
+class RefWithSha(t.NamedTuple):
+  ref: str
+  sha: str
+
+
+class Remote(t.NamedTuple):
+  name: str
+  fetch: str
+  push: str
 
 
 class Git:
 
   def __init__(self, cwd: str = None):
-    self.cwd = cwd
+    self.path = cwd or os.getcwd()
+
+  @property
+  @deprecated.deprecated('use Git.path instead', version='0.4.0')
+  def cwd(self) -> str:
+    return self.path
+
+  def check_call(self, command: t.List[str], stdout: t.Optional[int] = None) -> None:
+    sp.check_call(command, cwd=self.path, stdout=stdout)
+
+  def check_output(self, command: t.List[str], stderr: t.Optional[int] = None) -> bytes:
+    return sp.check_output(command, cwd=self.path, stderr=stderr)
+
+  def init(self) -> None:
+    self.check_call(['git', 'init', '.'])
 
   def clone(
       self,
-      to_directory: str,
       clone_url: str,
       branch: str = None,
       depth: int = None,
@@ -47,7 +78,7 @@ class Git:
       username: str = None,
       password: str = None,
       quiet: bool = False,
-  ) -> 'Git':
+  ) -> None:
     """
     Clone a Git repository to the *to_directory* from *clone_url*. If a relative path is
     specified in *to_directory*, it will be treated relative to the #Git.cwd path.
@@ -60,7 +91,7 @@ class Git:
       auth = ':'.join(t.cast(t.List[str], filter(bool, [username, password])))
       clone_url = schema + '://' + auth + '@' + remainder
 
-    command = ['git', 'clone', clone_url, to_directory]
+    command = ['git', 'clone', clone_url, self.path]
     if branch:
       command += ['-b', branch]
     if depth:
@@ -70,16 +101,18 @@ class Git:
     if quiet:
       command += ['-q']
 
-    sp.check_call(command, cwd=self.cwd)
-    return Git(to_directory)
+    # NOTE (NiklasRosenstein): We don't use #Git.check_call() as that would try to
+    # change directory to the clone target directory, which might not yet exist.
+    sp.check_call(command)
 
   def add(self, files: t.List[str]) -> None:
     """
     Add files to the index.
     """
 
+    assert isinstance(files, list), f'expected list, got {type(files).__name__}'
     command = ['git', 'add', '--'] + files
-    sp.check_call(command, cwd=self.cwd)
+    self.check_call(command)
 
   def get_branches(self) -> t.List[Branch]:
     """
@@ -88,7 +121,7 @@ class Git:
 
     command = ['git', 'branch']
     results = []
-    for line in sp.check_output(command, cwd=self.cwd).decode().splitlines():
+    for line in self.check_output(command).decode().splitlines():
       current = False
       if line.startswith('*'):
         line = line[1:]
@@ -118,7 +151,7 @@ class Git:
   def get_remote_refs(self, remote: str) -> t.List[RefWithSha]:
     result = []
     command = ['git', 'ls-remote', '--heads', 'origin']
-    for line in sp.check_output(command, cwd=self.cwd).decode().splitlines():
+    for line in self.check_output(command).decode().splitlines():
       sha, ref = line.split()
       result.append(RefWithSha(ref, sha))
     return result
@@ -127,6 +160,9 @@ class Git:
     refs = self.get_remote_refs(remote)
     return [x.ref[11:] for x in refs if x.ref.startswith('refs/heads/')]
 
+  def rename_branch(self, current: str, new: str) -> None:
+    self.check_call(['git', 'branch', '-m', current, new])
+
   def push(self, remote: str, *refs, force: bool = False) -> None:
     """
     Push the specified *refs* to the Git remote.
@@ -135,7 +171,7 @@ class Git:
     command = ['git', 'push', remote] + list(refs)
     if force:
       command.insert(2, '-f')
-    sp.check_call(command, cwd=self.cwd)
+    self.check_call(command)
 
   def pull(self, remote: str = None, branch: str = None, quiet: bool = False):
     """
@@ -145,19 +181,67 @@ class Git:
     command = ['git', 'pull']
     if remote and branch:
       command += [remote, branch]
-    if quiet:
-      command += ['-q']
     elif remote or branch:
       raise ValueError('remote and branch arguments can only be specified together')
+    if quiet:
+      command += ['-q']
 
-    sp.check_call(command, cwd=self.cwd)
+    self.check_call(command)
+
+  def fetch(
+    self,
+    remote: str = None,
+    all: bool = False,
+    tags: bool = False,
+    prune: bool = False,
+    prune_tags: bool = False,
+    argv: t.Optional[t.List[str]] = None
+  ) -> None:
+    """
+    Fetch a remote repository (or multiple).
+    """
+
+    command = ['git', 'fetch']
+    if remote:
+      command += [remote]
+    if all:
+      command += ['--all']
+    if tags:
+      command += ['--tags']
+    if prune:
+      command += ['--prune']
+    if prune_tags:
+      command += ['--prune-tags']
+    command += argv or []
+
+    self.check_call(command)
+
+  def remotes(self) -> t.List[Remote]:
+    """
+    List up all the remotes of the repository.
+    """
+
+    remotes: t.Dict[str, t.Dict[str, str]] = {}
+    for line in self.check_output(['git', 'remote', '-v']).decode().splitlines():
+      remote, url, kind = line.split()
+      remotes.setdefault(remote, {})[kind] = url
+
+    return [Remote(remote, urls['(fetch)'], urls['(push)']) for remote, urls in remotes.items()]
+
+  def add_remote(self, remote: str, url: str, argv: t.Optional[t.List[str]] = None) -> None:
+    """
+    Add a remote with the specified name.
+    """
+
+    command = ['git', 'remote', 'add', remote, url] + (argv or [])
+    self.check_call(command)
 
   def porcelain(self) -> t.Iterable[FileStatus]:
     """
     Returns the file status for the working tree.
     """
 
-    for line in sp.check_output(['git', 'status', '--porcelain'], cwd=self.cwd).decode().splitlines():
+    for line in self.check_output(['git', 'status', '--porcelain']).decode().splitlines():
       mode, filename = line.strip().partition(' ')[::2]
       yield FileStatus(mode, filename)
 
@@ -169,7 +253,7 @@ class Git:
     command = ['git', 'commit', '-m', message]
     if allow_empty:
       command.append('--allow-empty')
-    sp.check_call(command, cwd=self.cwd)
+    self.check_call(command)
 
   def tag(self, tag_name: str, force: bool = False) -> None:
     """
@@ -177,7 +261,7 @@ class Git:
     """
 
     command = ['git', 'tag', tag_name] + (['-f'] if force else [])
-    sp.check_call(command)
+    self.check_call(command)
 
   def rev_parse(self, rev: str) -> t.Optional[str]:
     """
@@ -186,7 +270,7 @@ class Git:
 
     command = ['git', 'rev-parse', rev]
     try:
-      return sp.check_output(command, stderr=sp.STDOUT, cwd=self.cwd).decode().strip()
+      return self.check_output(command, stderr=sp.STDOUT).decode().strip()
     except sp.CalledProcessError:
       return None
 
@@ -199,7 +283,7 @@ class Git:
     if path:
       command += ['--', path]
     try:
-      revlist = sp.check_output(command, stderr=sp.STDOUT, cwd=self.cwd).decode().strip().split('\n')
+      revlist = self.check_output(command, stderr=sp.STDOUT).decode().strip().split('\n')
     except sp.CalledProcessError:
       return []
     if revlist == ['']:
@@ -212,22 +296,29 @@ class Git:
     """
 
     try:
-      sp.check_call(['git', 'diff', '--exit-code'], stdout=sp.PIPE, cwd=self.cwd)
+      self.check_call(['git', 'diff', '--exit-code'], stdout=sp.PIPE)
       return False
     except sp.CalledProcessError as exc:
       if exc.returncode == 1:
         return True
       raise
 
-  def create_branch(self, name: str, orphan: bool = False) -> None:
+  def create_branch(self, name: str, orphan: bool = False, reset: bool = False, ref: t.Optional[str] = None) -> None:
     """
     Creates a branch.
     """
 
-    command = ['git', 'checkout', '-b', name]
+    command = ['git', 'checkout']
     if orphan:
-      command += ['--orphan']
-    sp.check_call(command, cwd=self.cwd)
+      if ref:
+        raise ValueError('cannot checkout orphan branch with ref')
+      command += ['--orphan', name]
+    else:
+      command += ['-B' if reset else '-b', name]
+      if ref:
+        command += [ref]
+
+    self.check_call(command)
 
   def checkout(self, ref: str = None, files: t.List[str] = None, quiet: bool = False):
     """
@@ -241,7 +332,7 @@ class Git:
       command += ['-q']
     if files:
       command += ['--'] + files
-    sp.check_call(command, cwd=self.cwd)
+    self.check_call(command)
 
   def reset(self, ref: str = None, files: t.List[str] = None, hard: bool = False, quiet: bool = False):
     """
@@ -255,14 +346,14 @@ class Git:
       command += ['-q']
     if files:
       command += ['--'] + files
-    sp.check_call(command, cwd=self.cwd)
+    self.check_call(command)
 
   def get_commit_message(self, rev: str) -> str:
     """
     Returns the commit message of the specified *rev*.
     """
 
-    return sp.check_output(['git', 'log', '-1', rev, '--pretty=%B']).decode()
+    return self.check_output(['git', 'log', '-1', rev, '--pretty=%B']).decode()
 
   def get_diff(self, files: t.List[str] = None, cached: bool = False):
     command = ['git', '--no-pager', 'diff', '--color=never']
@@ -270,7 +361,7 @@ class Git:
       command += ['--cached']
     if files is not None:
       command += ['--'] + files
-    return sp.check_output(command, cwd=self.cwd).decode()
+    return self.check_output(command).decode()
 
   def describe(self,
     all: bool = False,
@@ -290,7 +381,7 @@ class Git:
       command.append(commitish)
 
     try:
-      return sp.check_output(command, stderr=sp.DEVNULL).decode().strip()
+      return self.check_output(command, stderr=sp.DEVNULL).decode().strip()
     except FileNotFoundError:
       raise
     except sp.CalledProcessError:
